@@ -53,9 +53,15 @@ hookimpl = tox.hookimpl
 
 WITHIN_PROVISION = os.environ.get(str("TOX_PROVISION")) == "1"
 
+SUICIDE_TIMEOUT = 0.0
 INTERRUPT_TIMEOUT = 0.3
 TERMINATE_TIMEOUT = 0.2
 
+_FACTOR_LINE_PATTERN = re.compile(r"^([\w{}\.!,-]+)\:\s+(.+)")
+_ENVSTR_SPLIT_PATTERN = re.compile(r"((?:\{[^}]+\})+)|,")
+_ENVSTR_EXPAND_PATTERN = re.compile(r"\{([^}]+)\}")
+_WHITESPACE_PATTERN = re.compile(r"\s+")
+
 
 def get_plugin_manager(plugins=()):
     # initialize plugin manager
@@ -86,18 +92,21 @@ class Parser:
                 super(HelpFormatter, self).__init__(prog, max_help_position=35, width=190)
 
         self.argparser = argparse.ArgumentParser(
-            description="tox options", add_help=False, prog="tox", formatter_class=HelpFormatter,
+            description="tox options",
+            add_help=False,
+            prog="tox",
+            formatter_class=HelpFormatter,
         )
         self._testenv_attr = []
 
     def add_argument(self, *args, **kwargs):
-        """ add argument to command line parser.  This takes the
+        """add argument to command line parser.  This takes the
         same arguments that ``argparse.ArgumentParser.add_argument``.
         """
         return self.argparser.add_argument(*args, **kwargs)
 
     def add_testenv_attribute(self, name, type, help, default=None, postprocess=None):
-        """ add an ini-file variable for "testenv" section.
+        """add an ini-file variable for "testenv" section.
 
         Types are specified as strings like "bool", "line-list", "string", "argv", "path",
         "argvlist".
@@ -114,7 +123,7 @@ class Parser:
         self._testenv_attr.append(VenvAttribute(name, type, default, help, postprocess))
 
     def add_testenv_attribute_obj(self, obj):
-        """ add an ini-file variable as an object.
+        """add an ini-file variable as an object.
 
         This works as the ``add_testenv_attribute`` function but expects
         "name", "type", "help", and "postprocess" attributes on the object.
@@ -265,7 +274,10 @@ def parseconfig(args, plugins=()):
                 content = toml_content["tool"]["tox"]["legacy_tox_ini"]
             except KeyError:
                 continue
-        ParseIni(config, config_file, content)
+        try:
+            ParseIni(config, config_file, content)
+        except SkipThisIni:
+            continue
         pm.hook.tox_configure(config=config)  # post process config object
         break
     else:
@@ -276,6 +288,11 @@ def parseconfig(args, plugins=()):
         parser.parse_cli(args, strict=True)
         if option.help or option.helpini:
             return config
+        if option.devenv:
+            # To load defaults, we parse an empty config
+            ParseIni(config, py.path.local(), "")
+            pm.hook.tox_configure(config=config)
+            return config
         msg = "tox config file (either {}) not found"
         candidates = ", ".join(INFO.CONFIG_CANDIDATES)
         feedback(msg.format(candidates), sysexit=not (option.help or option.helpini))
@@ -298,7 +315,8 @@ def propose_configs(cli_config_file):
             from_folder = py.path.local(cli_config_file)
         else:
             print(
-                "ERROR: {} is neither file or directory".format(cli_config_file), file=sys.stderr,
+                "ERROR: {} is neither file or directory".format(cli_config_file),
+                file=sys.stderr,
             )
             return
     for basename in INFO.CONFIG_CANDIDATES:
@@ -319,7 +337,11 @@ def parse_cli(args, pm):
         raise SystemExit(0)
     interpreters = Interpreters(hook=pm.hook)
     config = Config(
-        pluginmanager=pm, option=option, interpreters=interpreters, parser=parser, args=args,
+        pluginmanager=pm,
+        option=option,
+        interpreters=interpreters,
+        parser=parser,
+        args=args,
     )
     return config, option
 
@@ -368,7 +390,9 @@ class SetenvDict(object):
                 return os.environ.get(name, default)
             self._lookupstack.append(name)
             try:
-                self.resolved[name] = res = self.reader._replace(val)
+                res = self.reader._replace(val)
+                res = res.replace("\\{", "{").replace("\\}", "}")
+                self.resolved[name] = res
             finally:
                 self._lookupstack.pop()
             return res
@@ -390,7 +414,9 @@ class SetenvDict(object):
 @tox.hookimpl
 def tox_addoption(parser):
     parser.add_argument(
-        "--version", action="store_true", help="report version information to stdout.",
+        "--version",
+        action="store_true",
+        help="report version information to stdout.",
     )
     parser.add_argument("-h", "--help", action="store_true", help="show help about options")
     parser.add_argument(
@@ -420,7 +446,9 @@ def tox_addoption(parser):
         help="show list of all defined environments (with description if verbose)",
     )
     parser.add_argument(
-        "-c", dest="configfile", help="config file name or directory with 'tox.ini' file.",
+        "-c",
+        dest="configfile",
+        help="config file name or directory with 'tox.ini' file.",
     )
     parser.add_argument(
         "-e",
@@ -439,7 +467,14 @@ def tox_addoption(parser):
     )
     parser.add_argument("--notest", action="store_true", help="skip invoking test commands.")
     parser.add_argument(
-        "--sdistonly", action="store_true", help="only perform the sdist packaging activity.",
+        "--sdistonly",
+        action="store_true",
+        help="only perform the sdist packaging activity.",
+    )
+    parser.add_argument(
+        "--skip-pkg-install",
+        action="store_true",
+        help="skip package installation for this run",
     )
     add_parallel_flags(parser)
     parser.add_argument(
@@ -476,7 +511,10 @@ def tox_addoption(parser):
         "(pip by default).",
     )
     parser.add_argument(
-        "-r", "--recreate", action="store_true", help="force recreation of virtual environments",
+        "-r",
+        "--recreate",
+        action="store_true",
+        help="force recreation of virtual environments",
     )
     parser.add_argument(
         "--result-json",
@@ -604,7 +642,9 @@ def tox_addoption(parser):
                     warnings.warn(
                         "conflicting basepython version (set {}, should be {}) for env '{}';"
                         "resolve conflict or set ignore_basepython_conflict".format(
-                            proposed_version, implied_version, testenv_config.envname,
+                            proposed_version,
+                            implied_version,
+                            testenv_config.envname,
                         ),
                     )
 
@@ -620,7 +660,7 @@ def tox_addoption(parser):
 
     def merge_description(testenv_config, value):
         """the reader by default joins generated description with new line,
-         replace new line with space"""
+        replace new line with space"""
         return value.replace("\n", " ")
 
     parser.add_testenv_attribute(
@@ -632,11 +672,17 @@ def tox_addoption(parser):
     )
 
     parser.add_testenv_attribute(
-        name="envtmpdir", type="path", default="{envdir}/tmp", help="venv temporary directory",
+        name="envtmpdir",
+        type="path",
+        default="{envdir}/tmp",
+        help="venv temporary directory",
     )
 
     parser.add_testenv_attribute(
-        name="envlogdir", type="path", default="{envdir}/log", help="venv log directory",
+        name="envlogdir",
+        type="path",
+        default="{envdir}/log",
+        help="venv log directory",
     )
 
     parser.add_testenv_attribute(
@@ -655,10 +701,14 @@ def tox_addoption(parser):
 
     parser.add_testenv_attribute_obj(PosargsOption())
 
+    def skip_install_default(testenv_config, value):
+        return value is True or testenv_config.config.option.skip_pkg_install is True
+
     parser.add_testenv_attribute(
         name="skip_install",
         type="bool",
         default=False,
+        postprocess=skip_install_default,
         help="Do not install the current package. This can be used when you need the virtualenv "
         "management but do not want to install the current package",
     )
@@ -749,6 +799,12 @@ def tox_addoption(parser):
     parser.add_testenv_attribute(
         name="whitelist_externals",
         type="line-list",
+        help="DEPRECATED: use allowlist_externals",
+    )
+
+    parser.add_testenv_attribute(
+        name="allowlist_externals",
+        type="line-list",
         help="each lines specifies a path or basename for which tox will not warn "
         "about it coming from outside the test environment.",
     )
@@ -827,6 +883,13 @@ def tox_addoption(parser):
 
     parser.add_testenv_attribute_obj(DepOption())
 
+    parser.add_testenv_attribute(
+        name="suicide_timeout",
+        type="float",
+        default=SUICIDE_TIMEOUT,
+        help="timeout to allow process to exit before sending SIGINT",
+    )
+
     parser.add_testenv_attribute(
         name="interrupt_timeout",
         type="float",
@@ -947,10 +1010,17 @@ class TestenvConfig:
 
     def get_envbindir(self):
         """Path to directory where scripts/binaries reside."""
-        if tox.INFO.IS_WIN and "jython" not in self.basepython and "pypy" not in self.basepython:
-            return self.envdir.join("Scripts")
-        else:
-            return self.envdir.join("bin")
+        is_bin = (
+            isinstance(self.python_info, NoInterpreterInfo)
+            or tox.INFO.IS_WIN is False
+            or self.python_info.implementation == "Jython"
+            or (
+                tox.INFO.IS_WIN
+                and self.python_info.implementation == "PyPy"
+                and self.python_info.extra_version_info < (7, 3, 1)
+            )
+        )
+        return self.envdir.join("bin" if is_bin else "Scripts")
 
     @property
     def envbindir(self):
@@ -1014,13 +1084,22 @@ def make_hashseed():
     return str(random.randint(1, max_seed))
 
 
+class SkipThisIni(Exception):
+    """Internal exception to indicate the parsed ini file should be skipped"""
+
+
 class ParseIni(object):
     def __init__(self, config, ini_path, ini_data):  # noqa
         config.toxinipath = ini_path
         using("tox.ini: {} (pid {})".format(config.toxinipath, os.getpid()))
-        config.toxinidir = config.toxinipath.dirpath()
+        config.toxinidir = config.toxinipath.dirpath() if ini_path.check(file=True) else ini_path
 
         self._cfg = py.iniconfig.IniConfig(config.toxinipath, ini_data)
+
+        if ini_path.basename == "setup.cfg" and "tox:tox" not in self._cfg:
+            verbosity1("Found no [tox:tox] section in setup.cfg, skipping.")
+            raise SkipThisIni()
+
         previous_line_of = self._cfg.lineof
 
         self.expand_section_names(self._cfg)
@@ -1041,7 +1120,10 @@ class ParseIni(object):
         context_name = getcontextname()
         if context_name == "jenkins":
             reader = SectionReader(
-                "tox:jenkins", self._cfg, prefix=prefix, fallbacksections=[fallbacksection],
+                "tox:jenkins",
+                self._cfg,
+                prefix=prefix,
+                fallbacksections=[fallbacksection],
             )
             dist_share_default = "{toxworkdir}/distshare"
         elif not context_name:
@@ -1075,8 +1157,9 @@ class ParseIni(object):
 
         reader.addsubstitutions(distdir=config.distdir)
         config.distshare = reader.getpath("distshare", dist_share_default)
-        config.temp_dir = reader.getpath("temp_dir", "{toxworkdir}/.tmp")
         reader.addsubstitutions(distshare=config.distshare)
+        config.temp_dir = reader.getpath("temp_dir", "{toxworkdir}/.tmp")
+        reader.addsubstitutions(temp_dir=config.temp_dir)
         config.sdistsrc = reader.getpath("sdistsrc", None)
         config.setupdir = reader.getpath("setupdir", "{toxinidir}")
         config.logdir = config.toxworkdir.join("log")
@@ -1196,7 +1279,10 @@ class ParseIni(object):
                 self._cfg.sections[section_name] = {}
             self._cfg.sections[section_name]["description"] = "meta tox"
             env_config = self.make_envconfig(
-                name, "{}{}".format(testenvprefix, name), reader._subs, config,
+                name,
+                "{}{}".format(testenvprefix, name),
+                reader._subs,
+                config,
             )
             env_config.deps = deps
             config.envconfigs[config.provision_tox_env] = env_config
@@ -1251,7 +1337,10 @@ class ParseIni(object):
             self._cfg.sections[section_name]["sitepackages"] = "False"
             self._cfg.sections[section_name]["description"] = "isolated packaging environment"
             config.envconfigs[name] = self.make_envconfig(
-                name, "{}{}".format(testenvprefix, name), reader._subs, config,
+                name,
+                "{}{}".format(testenvprefix, name),
+                reader._subs,
+                config,
             )
 
     def _list_section_factors(self, section):
@@ -1291,7 +1380,10 @@ class ParseIni(object):
                 elif atype == "basepython":
                     no_fallback = name in (config.provision_tox_env,)
                     res = reader.getstring(
-                        env_attr.name, env_attr.default, replace=replace, no_fallback=no_fallback,
+                        env_attr.name,
+                        env_attr.default,
+                        replace=replace,
+                        no_fallback=no_fallback,
                     )
                 elif atype == "space-separated-list":
                     res = reader.getlist(env_attr.name, sep=" ")
@@ -1371,7 +1463,7 @@ class ParseIni(object):
         The parser will see it as two different sections: [testenv:py36-cov], [testenv:py37-cov]
 
         """
-        factor_re = re.compile(r"\{\s*([\w\s,]+)\s*\}")
+        factor_re = re.compile(r"\{\s*([\w\s,-]+)\s*\}")
         split_re = re.compile(r"\s*,\s*")
         to_remove = set()
         for section in list(config.sections):
@@ -1423,12 +1515,12 @@ def _split_factor_expr_all(expr):
 
 def _expand_envstr(envstr):
     # split by commas not in groups
-    tokens = re.split(r"((?:\{[^}]+\})+)|,", envstr)
+    tokens = _ENVSTR_SPLIT_PATTERN.split(envstr)
     envlist = ["".join(g).strip() for k, g in itertools.groupby(tokens, key=bool) if k]
 
     def expand(env):
-        tokens = re.split(r"\{([^}]+)\}", env)
-        parts = [re.sub(r"\s+", "", token).split(",") for token in tokens]
+        tokens = _ENVSTR_EXPAND_PATTERN.split(env)
+        parts = [_WHITESPACE_PATTERN.sub("", token).split(",") for token in tokens]
         return ["".join(variant) for variant in itertools.product(*parts)]
 
     return mapcat(expand, envlist)
@@ -1514,13 +1606,21 @@ class SectionReader:
         if value is None or not replace:
             return default or {}
 
-        d = {}
+        env_values = {}
         for line in value.split(sep):
             if line.strip():
-                name, rest = line.split("=", 1)
-                d[name.strip()] = rest.strip()
-
-        return d
+                if line.startswith("#"):  # comment lines are ignored
+                    pass
+                elif line.startswith("file|"):  # file markers contain paths to env files
+                    file_path = line[5:].strip()
+                    if os.path.exists(file_path):
+                        with open(file_path, "rt") as file_handler:
+                            content = file_handler.read()
+                        env_values.update(self._getdict(content, "", sep, replace))
+                else:
+                    name, value = line.split("=", 1)
+                    env_values[name.strip()] = value.strip()
+        return env_values
 
     def getfloat(self, name, default=None, replace=True):
         s = self.getstring(name, default, replace=replace)
@@ -1592,7 +1692,7 @@ class SectionReader:
 
     def _apply_factors(self, s):
         def factor_line(line):
-            m = re.search(r"^([\w{}\.!,-]+)\:\s+(.+)", line)
+            m = _FACTOR_LINE_PATTERN.search(line)
             if not m:
                 return line
 
@@ -1720,7 +1820,10 @@ class Replacer:
                     )
                 x = str(cfg[section][item])
                 return self.reader._replace(
-                    x, name=item, section_name=section, crossonly=self.crossonly,
+                    x,
+                    name=item,
+                    section_name=section,
+                    crossonly=self.crossonly,
                 )
 
         raise tox.exception.ConfigError("substitution key {!r} not found".format(key))
@@ -1772,7 +1875,8 @@ class _ArgvlistReader:
             if current_command:
                 raise tox.exception.ConfigError(
                     "line-continuation ends nowhere while resolving for [{}] {}".format(
-                        reader.section_name, "commands",
+                        reader.section_name,
+                        "commands",
                     ),
                 )
         return commands
